package com.vulnit.projects.js;

import com.vulnit.projects.js.elements.JsComment;
import com.vulnit.projects.js.elements.JsElements;
import com.vulnit.projects.js.elements.JsString;
import com.vulnit.projects.js.elements.JsText;

/**
 * 
 * @author rmillet
 * 
 */
public final class SimpleJsParser {
	private final JsElements elements = new JsElements();
	private String input;
	private int pos = 0;

	/**
	 * 
	 * @param input
	 *            Js data to parse
	 */
	private SimpleJsParser(String input) {
		this.input = input;
	}

	/**
	 * Parse the Javascript
	 * 
	 * @return {@link JsElements} composing the tree
	 */
	private JsElements parse() {
		StringBuilder sb = new StringBuilder();
		while (!isEOF()) {
			char c = consume();
			if (c == '\'') {
				sb = addText(sb);
				parseSingleQuotedString();
			} else if (c == '"') {
				sb = addText(sb);
				parseDoubleQuotedString();
			} else if (c == '/' && current() == '/') {
				sb = addText(sb);
				parseSingleLineComment();
			} else if (c == '/' && current() == '*') {
				sb = addText(sb);
				parseMultiLineComment();
			} else {
				sb.append(c);
			}
		}
		sb = addText(sb);
		return elements;
	}

	/**
	 * Add a Text node
	 * 
	 * @param sb
	 *            The {@link StringBuilder}
	 * @return A new {@link StringBuilder}
	 */
	private StringBuilder addText(StringBuilder sb) {
		this.elements.add(new JsText(sb.toString()));
		return new StringBuilder();
	}

	/**
	 * Parse a SingleLine Comment ("// comment")
	 */
	private void parseSingleLineComment() {
		advance();
		StringBuilder sb = new StringBuilder();
		while (!isEOF()) {
			char c = consume();
			if (c == '\n' || c == '\r') {
				break;
			}
			sb.append(c);
		}
		this.elements.add(new JsComment(sb.toString(), false));
	}

	/**
	 * Parse a MultiLine Comment ("/* comment * /")
	 */
	private void parseMultiLineComment() {
		advance();
		StringBuilder sb = new StringBuilder();
		while (!isEOF()) {
			char c = consume();
			if (c == '*' && current() == '/') {
				break;
			}
			sb.append(c);
		}
		this.elements.add(new JsComment(sb.toString(), true));
		advance();
	}

	/**
	 * Parse a Double quoted String ("string")
	 */
	private void parseDoubleQuotedString() {
		StringBuilder sb = new StringBuilder();
		while (!isEOF()) {
			char c = consume();
			if (c == '"') {
				break;
			} else if (c == '\\') {
				if (current() != '"') {
					sb.append(c);
				}
				sb.append(consume());
			} else {
				sb.append(c);
			}
		}
		this.elements.add(new JsString(sb.toString(), '"'));
	}

	/**
	 * Parse a Single quoted String ('string')
	 */
	private void parseSingleQuotedString() {
		StringBuilder sb = new StringBuilder();
		while (!isEOF()) {
			char c = consume();
			if (c == '\'') {
				break;
			} else if (c == '\\') {
				if (current() != '\'') {
					sb.append(c);
				}
				sb.append(consume());
			} else {
				sb.append(c);
			}
		}
		this.elements.add(new JsString(sb.toString(), '\''));
	}

	/**
	 * Return the current element and advance
	 * 
	 * @return the current char
	 */
	private char consume() {
		char c = isEOF() ? '\0' : this.input.charAt(this.pos);
		++this.pos;
		return c;
	}

	/**
	 * 
	 * @return the current char
	 */
	private char current() {
		return isEOF() ? '\0' : this.input.charAt(this.pos);
	}

	/**
	 * 
	 * @return <code>true</code>When we are at the end of file
	 */
	private boolean isEOF() {
		return this.pos >= this.input.length();
	}

	/**
	 * Advance
	 */
	private void advance() {
		++this.pos;
	}

	/**
	 * Parse the javascript js
	 * 
	 * @param js
	 *            the javascript
	 * @return {@link JsElements} composing the tree
	 */
	public static JsElements parse(String js) {
		SimpleJsParser parser;
		if (js.startsWith("javascript:")) {
			parser = new SimpleJsParser(js.substring(11));
		} else {
			parser = new SimpleJsParser(js);
		}
		return parser.parse();
	}
}
