package com.vulnit.projects.xss;

import com.vulnit.projects.xss.plugins.Plugin;

import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.net.URISyntaxException;
import java.net.URLEncoder;
import java.util.ArrayList;
import java.util.List;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.GnuParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.OptionBuilder;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;
import org.jsoup.Jsoup;
import org.jsoup.nodes.Document;
import org.jsoup.select.Elements;

/**
 * 
 * @author rmillet
 * 
 */
public final class XssScanner {
	private static final String ERROR_TRIGGER = ")\"'&|";
	private final String websiteUrl;
	private final List<WebPage> webPages = new ArrayList<WebPage>();
	private PluginManager pluginManager;

	/**
	 * 
	 * @param entryUrl
	 *            The URL of the page to test
	 * @param postData
	 *            Post Data to send
	 * @throws URISyntaxException
	 *             When the entry point is not well-formatted
	 */
	public XssScanner(String entryUrl, String postData) throws URISyntaxException {
		WebPage webPage = new WebPage(entryUrl, postData);
		int pos = entryUrl.indexOf('/', 8); // search after "https://"
		if (pos >= 0) {
			this.websiteUrl = entryUrl.substring(0, pos + 1);
		} else {
			this.websiteUrl = entryUrl;
		}
		this.webPages.add(webPage);
		HttpClientManager.getInstance().setPort(webPage.getPort());
	}

	/**
	 * @return a list of {@link XssReport}
	 */
	public List<XssReport> scan() {
		List<XssReport> toReport = new ArrayList<XssReport>();

		this.pluginManager = new PluginManager();

		for (WebPage webPage : this.webPages) {
			while (webPage.hasNextParameter()) {
				try {
					boolean needErrorTrigger = false;
					String searchText = Plugin.getRandomTag();
					Document doc = Jsoup.parse(webPage.getContent(searchText));
					Elements elements = doc.getElementsContaining(searchText);
					if (elements.isEmpty()) {
						// Try to trigger errors (the input may be displayed in error messages)
						searchText = ERROR_TRIGGER + searchText;
						doc = Jsoup.parse(webPage.getContent(searchText));
						elements = doc.getElementsContaining(searchText);
						needErrorTrigger = true;
					}
					// Make new tags to not find previous injected payloads.
					Plugin.generateTags();
					String successfulPayload = pluginManager.exec(webPage, elements, searchText,
							needErrorTrigger ? ERROR_TRIGGER : "");
					if (successfulPayload != null) {
						toReport.add(new XssReport(successfulPayload, webPage.getAttributeName(), webPage
								.toString(successfulPayload)));
					}
				} catch (IOException e) {
					System.err.println(e.getMessage());
				}
			}
		}
		return toReport;
	}

	/**
	 * Add a new WebPage to scan if the url is in the entry website
	 * 
	 * @param url
	 *            The new URL
	 * @param postData
	 *            The post data
	 */
	public void addWebPage(String url, String postData) {
		try {
			if (url.startsWith(this.websiteUrl)) {
				WebPage webPage = new WebPage(url, postData);
				if (!this.webPages.contains(webPage)) {
					this.webPages.add(webPage);
				}
			}
		} catch (URISyntaxException e) {
			e.printStackTrace();
		}
	}

	/**
	 * Set options
	 * 
	 * @param key
	 *            Option name
	 * @param value
	 *            Option value
	 */
	private void setOption(String key, String value) {
		if (value == null) {
			return;
		}

		if (key.equals("cookies")) {
			this.setCookies(value);
		} else if (key.equals("referer")) {
			this.setReferer(value);
		} else if (key.equals("user-agent")) {
			this.setUserAgent(value);
		} else if (key.equals("timeout")) {
			this.setTimeout(Integer.parseInt(value) * 1000);
		} else if (key.equals("string-tag")) {
			this.setTag(value);
		} else if (key.equals("num-tag")) {
			this.setNumTag(value);
		}
	}

	/**
	 * Set the cookies (cookie1=value1;cookie2=value2;)
	 * 
	 * @param cookies
	 *            The cookies
	 */
	public void setCookies(String cookies) {
		HttpClientManager.getInstance().setCookies(cookies);

	}

	/**
	 * Set the referer
	 * 
	 * @param referer
	 *            The referer
	 */
	public void setReferer(String referer) {
		HttpClientManager.getInstance().setReferer(referer);
	}

	/**
	 * Set the user agent
	 * 
	 * @param userAgent
	 *            The user agent
	 */
	public void setUserAgent(String userAgent) {
		HttpClientManager.getInstance().setUserAgent(userAgent);
	}

	/**
	 * Set the timeout
	 * 
	 * @param timeout
	 *            The timeout (milliseconds)
	 */
	public void setTimeout(int timeout) {
		HttpClientManager.getInstance().setTimeout(timeout);
	}

	/**
	 * Set the string tag (used to identify payloads)
	 * 
	 * @param tag
	 *            The tag
	 */
	public void setTag(String tag) {
		Plugin.setTag(tag);
	}

	/**
	 * Set the numeric tag (used to identify payloads)
	 * 
	 * @param numtag
	 *            The tag
	 */
	public void setNumTag(String numtag) {
		Plugin.setNumTag(numtag);
	}

	/**
	 * 
	 * @param usernamePassword
	 *            auth-cred
	 * @param type
	 *            auth-type (Basic, Digest or NTLM)
	 * @throws ParseException
	 *             If one of them is null
	 */
	public void setCredentials(String usernamePassword, String type) throws ParseException {
		if (usernamePassword == null && type == null) {
			return;
		}
		if ((usernamePassword == null && !"spnego".equalsIgnoreCase(type)) || type == null) {
			throw new ParseException("Both auth-type and auth-cred must be set for type '"
					+ type.toLowerCase() + "'");
		}
		HttpClientManager.getInstance().setCredentials(usernamePassword, type);
	}

	/**
	 * Display the help
	 * 
	 * @param options
	 *            The command line options
	 */
	private static void printHelp(Options options) {
		String version = XssScanner.class.getPackage() != null ? XssScanner.class.getPackage()
				.getImplementationVersion() : null;
				HelpFormatter formatter = new HelpFormatter();
		formatter.printHelp("java -jar xss-scanner-" + version + ".jar", options, true);
		System.exit(1);
	}

	/**
	 * @param args
	 *            Command line arguments
	 */
	@SuppressWarnings("static-access")
	public static void main(String[] args) {
		Options options = new Options();

		options.addOption("h", "help", false, "Print this message");
		options.addOption("v", "version", false, "Print the current version");

		// Options for url
		Option url = OptionBuilder.withArgName("url").withLongOpt("url").hasOptionalArgs().withDescription(
				"URLs to test").create("u");
		options.addOption(url);
		options.addOption("p", "post-data", true, "Post-Data (eg: login=test%20login&password=)");

		// Options for HttpClient
		options.addOption("c", "cookies", true, "Cookies (eg: cookie1=value1;cookie2=value2;)");
		options.addOption("r", "referer", true, "Referer");
		options.addOption("a", "user-agent", true, "User-agent");
		options.addOption("t", "timeout", true, "Timeout");

		// Options for manual tests
		options.addOption("s", "string-tag", true, "String to inject (default: VulnIT)");
		options
				.addOption("n", "num-tag", true,
						"Number or predefined javascript variable to inject (eg: 42 or document.title ...). (default: 33155949271)");

		Option authCred = OptionBuilder.withLongOpt("auth-cred").hasArg().withDescription(
				"Authentication credentials (username:password or domain/username:password").create();
		options.addOption(authCred);
		Option authType = OptionBuilder.withLongOpt("auth-type").hasArg().withDescription(
				"Authentication type (Basic, Digest, NTLM, SPNEGO)").create();
		options.addOption(authType);

		GnuParser gnuParser = new GnuParser();
		try {
			CommandLine commandLine = gnuParser.parse(options, args);
			if (commandLine.hasOption("help")) {
				printHelp(options);
			} else if (commandLine.hasOption("version")) {
				String version = XssScanner.class.getPackage() != null ? XssScanner.class.getPackage()
						.getImplementationVersion() : null;
				System.out.println("Version: " + version);
				System.exit(0);
			} else if (commandLine.getOptionValues("url") == null) {
				throw new ParseException("Missing argument for option: url");
			} else if (commandLine.getOptionValues("url").length > 1 && commandLine.hasOption("post-data")) {
				throw new ParseException("Can't have more that one URL with post-data argument");
			}

			XssScanner scanner = new XssScanner(commandLine.getOptionValue("url"), commandLine
					.getOptionValue("post-data"));

			scanner.setOption("cookies", commandLine.getOptionValue("cookies"));
			scanner.setOption("referer", commandLine.getOptionValue("referer"));
			scanner.setOption("user-agent", commandLine.getOptionValue("user-agent"));

			scanner.setOption("string-tag", commandLine.getOptionValue("string-tag"));
			scanner.setOption("num-tag", commandLine.getOptionValue("num-tag"));

			scanner.setOption("timeout", commandLine.getOptionValue("timeout"));

			scanner.setCredentials(commandLine.getOptionValue("auth-cred"), commandLine
					.getOptionValue("auth-type"));

			String[] urls = commandLine.getOptionValues("url");
			for (int i = 1; i < urls.length; ++i) {
				scanner.addWebPage(urls[i], null);
			}

			List<XssReport> reports = scanner.scan();
			if (reports.isEmpty()) {
				System.out.println("=== No XSS found ===");
			} else {
				System.out.println("=== XSS found ===");
				for (XssReport report : reports) {
					// Can be parsed automatically as all but payload is URL encoded
					try {
						System.out.println(report.getAttributeName() + " "
								+ URLEncoder.encode(report.getPayload(), "UTF-8") + " " + report.toString());
					} catch (UnsupportedEncodingException e) {
						e.printStackTrace();
					}
				}
			}

		} catch (ParseException e) {
			System.err.println(e.getMessage());
			System.err.flush();
			printHelp(options);
		} catch (URISyntaxException e) {
			System.err.println(e.getMessage());
		}
	}
}
